local jbeam_io = require('jbeam/io')
local json = require("json")

local M = {}

local tintable_parts_list = {}
local allTintablePartsEquipped = false
local supportsTintable = false
local anyTintableEquipped = false

local glass_parts = {
    doorglass = {
        slot = "doorglass",
        name = "doorglass"
    },
    windshield = {
        slot = "windshield",
        name = "windshield"
    },
    backlight = {
        slot = "backlight",
        name = "backlight"
    },
    sunroof = {
        slot = "sunroof",
        name = "sunroof"
    },
    quarterglass = {
        slot = "quarterglass",
        name = "quarterglass"
    },
    sideglass = {
        slot = "sideglass",
        name = "sideglass"
    },
    tailgateglass = {
        slot = "tailgateglass",
        name = "tailgateglass"
    },
    -- window = {
    --     slot = "window",
    --     name = "window"
    -- },
    backglass = {
        slot = "backglass",
        name = "backglass"
    },
    glass_F = {
        slot = "glass_F",
        name = "windshield"
    },
    glass_R = {
        slot = "glass_R",
        name = "backlight"
    },
    roofglass = {
        slot = "roofglass",
        name = "roofglass"
    },
    rearglass = {
        slot = "rearglass",
        name = "rearglass"
    },
    enginelid = {
        slot = "enginelid",
        name = "enginelid"
    },
    trunklidglass = {
        slot = "rearglass",
        name = "trunklidglass"
    },
    -- bluebuck_glass = {
    --     slot = "glass",
    --     name = "doorglass"
    -- },
    bluebuck_doorglass_FL = {
        slot = "_glass_FL",
        name = "doorglass_FL"
    },
    bluebuck_doorglass_FR = {
        slot = "_glass_FR",
        name = "doorglass_FR"
    },
    bluebuck_doorglass_RL = {
        slot = "_glass_RL",
        name = "doorglass_RL"
    },
    bluebuck_doorglass_RR = {
        slot = "_glass_RR",
        name = "doorglass_RR"
    },
    bluebuck_windowframe = {
        slot = "windowframe",
        name = "windowframe"
    }
}

local function addTintableParts()
    local veh = be:getPlayerVehicle(0)
    local veh_data = extensions.core_vehicle_manager.getPlayerVehicleData()
    local veh_name = veh:getJBeamFilename()

    local all_slots = jbeam_io.getAvailableSlotMap(veh_data.ioCtx)
    local chosen_parts = veh_data.chosenParts

    local added_part = false

    tintable_parts_list = {}
    supportsTintable = false

    for key, glass_part in pairs(glass_parts) do
        local slot_to_find = glass_part.slot
        local name_to_find = glass_part.name

        for slot_name, curr_part in pairs(chosen_parts) do
            local parts_for_slot = all_slots[slot_name]

            if parts_for_slot then
                if string.find(slot_name, slot_to_find) then
                    for _, part_name in pairs(parts_for_slot) do
                        if string.find(part_name, "Jamie599s_" .. veh_name .. "_" .. name_to_find) then
                            supportsTintable = true
                            if chosen_parts[slot_name] ~= part_name then
                                supportsTintable = true
                                chosen_parts[slot_name] = part_name
                                added_part = true
                            end
                        end
                    end
                end
            end
        end
    end

    -- If parts were added, set parts config and reset vehicle
    if added_part then
        extensions.core_vehicle_partmgmt.setPartsConfig(chosen_parts, true)
        allTintablePartsEquipped = true
    end

end

local function tintablePartsToAdd()
    local veh = be:getPlayerVehicle(0)
    local veh_data = extensions.core_vehicle_manager.getPlayerVehicleData()
    local veh_name = veh:getJBeamFilename()

    local all_slots = jbeam_io.getAvailableSlotMap(veh_data.ioCtx)
    local chosen_parts = veh_data.chosenParts

    local added_part = false

    tintable_parts_list = {}
    allTintablePartsEquipped = true
    supportsTintable = false

    for key, glass_part in pairs(glass_parts) do
        local slot_to_find = glass_part.slot
        local name_to_find = glass_part.name

        for slot_name, curr_part in pairs(chosen_parts) do
            local parts_for_slot = all_slots[slot_name]

            if parts_for_slot then
                if string.find(slot_name, slot_to_find) then
                    for _, part_name in pairs(parts_for_slot) do
                        if string.find(part_name, "Jamie599s_" .. veh_name .. "_" .. name_to_find) then
                            supportsTintable = true
                            if chosen_parts[slot_name] ~= part_name then
                                allTintablePartsEquipped = false
                            end
                        end
                    end
                end
            end
        end
    end

    return allTintablePartsEquipped

end

local function vehSupportsTintable()
    -- -- print("supportsTintable: " .. tostring(supportsTintable))
    return supportsTintable
end

local function anyTintableEquipped()
    return anyTintableEquipped
end

local function getTintableParts()

    local veh = be:getPlayerVehicle(0)
    local veh_data = extensions.core_vehicle_manager.getPlayerVehicleData()
    local veh_name = veh:getJBeamFilename()

    local chosen_parts = veh_data.chosenParts

    tintable_parts_list = {}
    anyTintableEquipped = false

    for key, glass_part in pairs(glass_parts) do
        local slot_to_find = glass_part.slot
        local name_to_find = glass_part.name

        for slot_name, curr_part in pairs(chosen_parts) do

            if string.find(curr_part, "Jamie599s_" .. veh_name .. "_" .. name_to_find) then
                if string.find(slot_name, slot_to_find) then
                    ---- -- print("")
                    ---- -- print("Found tintable part: " .. curr_part)
                    ---- -- print("Slot: " .. slot_name)
                    ---- -- print("Name to find: " .. name_to_find)
                    ---- -- print("Slot to find: " .. slot_to_find)
                    ---- -- print("")

                    local category = "Glass"
                    local name = slot_name:gsub(veh_name .. "_", ""):gsub("_", " ")
                    anyTintableEquipped = true

                    if string.find(slot_name, "doorglass") then
                        category = "Door Glass"
                        if string.find(name, "FL") then
                            name = "Front Left"
                        elseif string.find(name, "FR") then
                            name = "Front Right"
                        elseif string.find(name, "RL") then
                            name = "Rear Left"
                        elseif string.find(name, "RR") then
                            name = "Rear Right"
                        elseif string.find(name, "L") then
                            name = "Left"
                        elseif string.find(name, "R") then
                            name = "Right"
                        end

                    elseif string.find(slot_name, "quarterglass") then
                        category = "Quarter Glass"
                        if string.find(name, "FL") then
                            name = "Front Left"
                        elseif string.find(name, "FR") then
                            name = "Front Right"
                        elseif string.find(name, "RL") then
                            name = "Rear Left"
                        elseif string.find(name, "RR") then
                            name = "Rear Right"
                        elseif string.find(name, "L") then
                            name = "Left"
                        elseif string.find(name, "R") then
                            name = "Right"
                        end

                    elseif string.find(slot_name, "glass_FL") or string.find(slot_name, "glass_FR") or
                        string.find(slot_name, "glass_RL") or string.find(slot_name, "glass_RR") then
                        category = "Door Glass"
                        if string.find(name, "FL") then
                            name = "Front Left"
                        elseif string.find(name, "FR") then
                            name = "Front Right"
                        elseif string.find(name, "RL") then
                            name = "Rear Left"
                        elseif string.find(name, "RR") then
                            name = "Rear Right"
                        elseif string.find(name, "L") then
                            name = "Left"
                        elseif string.find(name, "R") then
                            name = "Right"
                        end
                        
                        if string.find(curr_part, "bluebuck_doorglass_FL_4door") then
                            table.insert(tintable_parts_list, {
                                slot = slot_name,
                                part = "Jamie599s_bluebuck_quarterglass_FL_sedan_tintable",
                                name = "Left",
                                category = "Quarter Glass",
                                tint = 100 - M.round(M.CalcOutsideTint(M.getTint("Jamie599s_bluebuck_quarterglass_FL_sedan_tintable"), 1))
                            })
                        elseif string.find(curr_part, "bluebuck_doorglass_FR_4door") then
                            table.insert(tintable_parts_list, {
                                slot = slot_name,
                                part = "Jamie599s_bluebuck_quarterglass_FR_sedan_tintable",
                                name = "Right",
                                category = "Quarter Glass",
                                tint = 100 - M.round(M.CalcOutsideTint(M.getTint("Jamie599s_bluebuck_quarterglass_FR_sedan_tintable"), 1))
                            })
                        end

                    else
                        if string.find(curr_part, "windowframe_FL_2door_hardtop") then
                            curr_part = "Jamie599s_" .. veh_name .. "_quarterglass_FR_hardtop_tintable"
                            category = "Quarter Glass"
                            name = "Right"
                        elseif string.find(curr_part, "windowframe_FR_2door_hardtop") then
                            category = "Quarter Glass"
                            curr_part = "Jamie599s_" .. veh_name .. "_quarterglass_FL_hardtop_tintable"
                            name = "Left"
                        end

                        if string.find(name, "windshield") or string.find(name, "glass F") then
                            name = "Windshield"
                        elseif string.find(name, "tailgate") then
                            name = "Tailgate"
                        elseif string.find(name, "backlight") or string.find(name, "glass R") then
                            name = "Backlight"
                        elseif string.find(name, "sunroof") or string.find(name, "roofglass") then
                            name = "Sunroof"
                        elseif string.find(name, "rearglass") then
                            name = "Rear"
                        elseif string.find(name, "enginelid") then
                            name = "Engine Lid"
                        end
                    end

                    table.insert(tintable_parts_list, {
                        slot = slot_name,
                        part = curr_part,
                        name = name,
                        category = category,
                        tint = 100 - M.round(M.CalcOutsideTint(M.getTint(curr_part), 1))
                    })

                end
            end
        end
    end
    return tintable_parts_list
end

local function setTint(materialName, value, n)
    local material = null
    value = 100 - value
    if type(materialName) == "string" then
        -- print("materialvalue: " .. n)
        if n == 0 then
            value = M.CalcOutsideTint(value, 0)
        elseif n == 1 then
            value = M.CalcInsideTint(value, 0)
            materialName = materialName:gsub("_tintable", "_int_tintable")
        elseif n == 2 then
            value = M.CalcOutsideTint(value, 0)
            materialName = materialName:gsub("_tintable", "_dmg_tintable")
        end

        material = scenetree.findObject(materialName)

        if material then
            if material.___type == "class<Material>" then

                if n == 0 then
                    if material.activeLayers < 2 then
                        material:setField("activeLayers", 0, 2)
                    end
                    -- if material:getField("opacityMap", 0) ~= "" then
                    --    -- -- print("opacityMap: " .. material:getField("opacityMap", 0))
                    --     material:setField("opacityMap", 0, "")
                    -- end
                    -- if material:getField("opacityFactor", 0) ~= "0.5" then
                    --    -- -- print("opacityFactor: " .. material:getField("opacityFactor", 0))
                    --     material:setField("opacityFactor", 0, "0.5")
                    -- end

                    material:setField("baseColorFactor", 0,
                        1 - (value / 100) .. " " .. 1 - (value / 100) .. " " .. 1 - (value / 100) .. " 1")
                    material:setField("opacityFactor", 1, value / 100)

                    if material:getField("baseColorFactor", 1) ~= "0 0 0 1" then
                        ---- -- print("baseColorFactor: " .. material:getField("baseColorFactor", 1))
                        material:setField("baseColorFactor", 1, "0 0 0 1")
                    end
                    if material:getField("roughnessFactor", 1) ~= "0" then
                        ---- -- print("roughnessFactor: " .. material:getField("roughnessFactor", 1))
                        material:setField("roughnessFactor", 1, 0)
                    end
                    material:flush()
                    material:reload()
                    ---- -- print("Set " .. "opacityFactor" .. " to " .. value .. " on " .. material:getName())
                    return true
                elseif n == 1 then
                    if material.activeLayers < 2 then
                        material:setField("activeLayers", 0, 2)
                    end
                    ---- -- print(material:getField("translucentBlendOp", 0))
                    ---- -- print(material:getField("alphaRef", 0))
                    ---- -- print(material:getField("alphaTest", 0))

                    local material_avaliable = false

                    material:getField("version", 0)

                    material:setField("version", 0, "1.5")
                    -- material:setField("translucent", 0, "1")
                    -- material:setField("alphaRef", 0, 50)
                    material:setField("alphaTest", 0, "0")
                    material:setField("translucent", 0, "1")
                    material:setField("translucentBlendOp", 0, "PreMulAlpha")

                    ---- -- print(material:getField("translucentBlendOp", 0))
                    ---- -- print(material:getField("alphaRef", 0))
                    ---- -- print(material:getField("alphaTest", 0))

                    ---- -- print(value / 100)

                    material:setField("opacityMap", 0, material:getField("colorMap", 0))
                    if string.len(material:getField("opacityMap", 0)) > 5 then
                        material:setField("opacityFactor", 0, "1")
                    else
                        material:setField("opacityFactor", 0, "0.5")
                    end
                    -- material:setField("baseColorFactor", 0,
                    --     1 - (value / 100) .. " " .. 1 - (value / 100) .. " " .. 1 - (value / 100) .. " 1")
                    material:setField("baseColorFactor", 0, "0 0 0 1")
                    material:setField("opacityFactor", 1, value / 100)
                    material:setField("baseColorFactor", 1, "0 0 0 1")
                    material:setField("roughnessFactor", 0, 1)
                    material:setField("roughnessFactor", 1, 1)

                    material:flush()
                    material:reload()
                    ---- -- print("Set " .. "opacityFactor" .. " to " .. value / 100 .. " on " .. material:getName())
                    return true
                elseif n == 2 then
                    -- print("materialName: " .. materialName)
                    if string.find(materialName, "windshield") then
                        -- print("windshield")
                        if material.activeLayers < 2 then
                            material:setField("activeLayers", 0, 2)
                        end

                        material:setField("alphaTest", 0, "0")
                        material:setField("translucent", 0, "1")
                        material:setField("translucentBlendOp", 0, "PreMulAlpha")

                        material:setField("opacityFactor", 1, value / 100)
                        material:setField("baseColorFactor", 1, "0 0 0 1")
                        material:setField("roughnessFactor", 1, 0)
                    else
                        -- print("not windshield")
                        if material.activeLayers > 1 then
                            material:setField("activeLayers", 0, 1)
                        end
                    end
                    material:setField("baseColorFactor", 0,
                        1 - (value / 100) .. " " .. 1 - (value / 100) .. " " .. 1 - (value / 100) .. " 1")

                    material:flush()
                    material:reload()
                    return true
                end
            else
                log('E', "", "Given object is not a material.")
                -- print("Given object is not a material.")
                -- print("Material " .. material .. " not found.")
                return false
            end
        else
            log('E', "", "Given object is not a material.")
            -- print("Given object is not a material.")
            -- print("Material " .. material .. " not found.")
            return false
        end
    end
end

local function getTint(material)

    if type(material) == "string" then
        material = scenetree.findObject(material)
    end

    -- dump(material)

    if material.___type == "class<Material>" then
        if material.activeLayers < 2 then
            return 0
        else
            local value = material:getField("opacityFactor", 1)
            -- -- print("Got " .. "opacityFactor" .. " of " .. value .. " on " .. material:getName())
            return value * 100
        end
    else
        log('E', "", "Given object is not a material.")
        -- -- print("Given object is not a material.")
        return null
    end
end

local function round(num)
    return num + (2 ^ 52 + 2 ^ 51) - (2 ^ 52 + 2 ^ 51)
end

local function CalcOutsideTint(x, n)
    local a = -0.000115049
    local b = 0.00942252
    local c = -0.240838
    local d = 2.58118
    local e = 0

    if n == 0 then
        return M.fourthOrderPoly(x / 10, a, b, c, d, e) * 10
    else
        return M.solveFourthOrderPoly(x / 10, a, b, c, d, e) * 10
    end
end

local function CalcInsideTint(x, n)
    local a = -0.000243134
    local b = 0.0165652
    local c = -0.290157
    local d = 2.48818
    local e = 0

    if n == 0 then
        return M.fourthOrderPoly(x / 10, a, b, c, d, e) * 10
    else
        return M.solveFourthOrderPoly(x / 10, a, b, c, d, e) * 10
    end
end

local function fourthOrderPoly(x, a, b, c, d, e)
    return (a * x ^ 4 + b * x ^ 3 + c * x ^ 2 + d * x + e)
end

-- solve the 4th Order Polynomial equation for y
local function solveFourthOrderPoly(y, a, b, c, d, e)

    local function f(x)
        return a * x ^ 4 + b * x ^ 3 + c * x ^ 2 + d * x + e - y
    end

    -- Use Newton's method to find the root of the equation
    local eps = 1e-6 -- tolerance for convergence
    local maxIter = 100 -- maximum number of iterations
    local x0 = 0 -- initial guess for the root
    local x = x0

    for i = 1, maxIter do
        local fx = f(x)
        local dfx = 4 * a * x ^ 3 + 3 * b * x ^ 2 + 2 * c * x + d
        local dx = fx / dfx

        if math.abs(dx) < eps then
            return x
        end

        x = x - dx
    end

    -- If the root was not found, return nil
    return nil
end

M.addTintableParts = addTintableParts
M.getTintableParts = getTintableParts
M.tintablePartsToAdd = tintablePartsToAdd
M.vehSupportsTintable = vehSupportsTintable
M.anyTintableEquipped = anyTintableEquipped
M.setTint = setTint
M.getTint = getTint
M.round = round
M.fourthOrderPoly = fourthOrderPoly
M.solveFourthOrderPoly = solveFourthOrderPoly
M.CalcOutsideTint = CalcOutsideTint
M.CalcInsideTint = CalcInsideTint

return M
